from std_iostream cimport istream, ostream
from trie cimport Trie

cdef extern from "<marisa/iostream.h>" namespace "marisa" nogil:

    istream &read(istream &stream, Trie *trie)
    ostream &write(ostream &stream, Trie &trie)



import io
import pickle

import pytest
import hypothesis.strategies as st
from hypothesis import assume, given

import marisa_trie

from .utils import text


@given(st.sets(text), st.lists(st.binary()), text)
def test_contains(keys, values, missing_key):
    assume(missing_key not in keys)

    data = zip(keys, values)
    trie = marisa_trie.BytesTrie(data)

    for word, value in data:
        assert word in trie

    assert missing_key not in trie


@given(st.sets(text), st.lists(st.binary()), text)
def test_getitem(keys, values, missing_key):
    assume(missing_key not in keys)

    data = zip(keys, values)
    trie = marisa_trie.BytesTrie(data)

    for key, value in data:
        assert trie[key] == [value]

    with pytest.raises(KeyError):
        trie[missing_key]


@pytest.mark.parametrize("data", [[], [("foo", b"bar")]])
def test_getitem_missing(data):
    trie = marisa_trie.BytesTrie(data)
    with pytest.raises(KeyError):
        trie["missing"]


def test_getitem_multiple():
    data = [
        ("foo", b"x"),
        ("fo", b"y"),
        ("foo", b"a"),
    ]
    trie = marisa_trie.BytesTrie(data)
    assert trie["fo"] == [b"y"]
    assert trie["foo"] == [b"a", b"x"]


def test_null_bytes_in_values():
    data = [("foo", b"\x00\x00bar\x00")]
    trie = marisa_trie.BytesTrie(data)

    for key, value in data:
        assert trie[key] == [value]


def test_items():
    data = [
        ("fo", b"y"),
        ("foo", b"x"),
        ("foo", b"a"),
    ]
    trie = marisa_trie.BytesTrie(data)
    assert set(trie.items()) == set(data)
    assert set(trie.items("f")) == set(data)
    assert set(trie.items("fo")) == set(data)
    assert set(trie.items("foo")) == set(data[1:])
    assert trie.items("food") == []
    assert trie.items("bar") == []


@given(st.sets(text), st.lists(st.binary()))
def test_iteritems(keys, values):
    trie = marisa_trie.BytesTrie(zip(keys, values))
    assert trie.items() == list(trie.iteritems())

    for key in keys:
        prefix = key[:5]
        assert trie.items(prefix) == list(trie.iteritems(prefix))


def test_keys():
    trie = marisa_trie.BytesTrie(
        [
            ("foo", b"x"),
            ("fo", b"y"),
            ("foo", b"a"),
        ]
    )

    # FIXME: ordering?
    assert trie.keys() == ["foo", "foo", "fo"]
    assert trie.keys("f") == ["foo", "foo", "fo"]
    assert trie.keys("fo") == ["foo", "foo", "fo"]
    assert trie.keys("foo") == ["foo", "foo"]
    assert trie.keys("food") == []
    assert trie.keys("bar") == []


@given(st.sets(text), st.lists(st.binary()))
def test_iterkeys(keys, values):
    trie = marisa_trie.BytesTrie(zip(keys, values))
    assert trie.keys() == list(trie.iterkeys())

    for key in keys:
        prefix = key[:5]
        assert trie.keys(prefix) == list(trie.iterkeys(prefix))


@given(st.sets(st.tuples(text, st.binary())))
def test_dumps_loads(data):
    trie = marisa_trie.BytesTrie(data)

    buf = io.BytesIO()
    pickle.dump(trie, buf)
    buf.seek(0)

    assert trie == pickle.load(buf)



"""
Shamelessly inspired from https://github.com/pypa/twine/blob/main/twine/commands/check.py
"""
import io
import re
from importlib.metadata import metadata

from readme_renderer.rst import render

# Regular expression used to capture and reformat docutils warnings into
# something that a human can understand. This is loosely borrowed from
# Sphinx: https://github.com/sphinx-doc/sphinx/blob
# /c35eb6fade7a3b4a6de4183d1dd4196f04a5edaf/sphinx/util/docutils.py#L199
_REPORT_RE = re.compile(
    r"^<string>:(?P<line>(?:\d+)?): "
    r"\((?P<level>DEBUG|INFO|WARNING|ERROR|SEVERE)/(\d+)?\) "
    r"(?P<message>.*)",
    re.DOTALL | re.MULTILINE,
)


class _WarningStream:
    def __init__(self) -> None:
        self.output = io.StringIO()

    def write(self, text: str) -> None:
        matched = _REPORT_RE.search(text)

        if not matched:
            self.output.write(text)
            return

        self.output.write(
            "line {line}: {level_text}: {message}\n".format(
                level_text=matched.group("level").capitalize(),
                line=matched.group("line"),
                message=matched.group("message").rstrip("\r\n"),
            )
        )

    def __str__(self) -> str:
        return self.output.getvalue()


def test_check_pypi_rendering():
    lines = metadata("marisa-trie").get("Summary", "").splitlines()
    description = lines.pop(0) + "\n"
    description += "\n".join(l[8:] for l in lines)

    warnings = _WarningStream()
    rendering = render(description, stream=warnings)
    print(description)
    print(warnings)
    assert not str(warnings)
    assert rendering is not None



import gzip
import os
import random
import string
import struct
import timeit

import marisa_trie


def words100k():
    zip_name = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "words100k.txt.gz"
    )

    return list(map(str.rstrip, gzip.open(zip_name, "rt")))


def random_words(num):
    russian = "абвгдеёжзиклмнопрстуфхцчъыьэюя"
    alphabet = f"{russian}{string.ascii_letters}"
    return [
        "".join(random.choice(alphabet) for _ in range(random.randint(1, 15)))
        for _ in range(num)
    ]


def truncated_words(words):
    return [word[:3] for word in words]


def prefixes1k(words, prefix_len):
    words = [w for w in words if len(w) >= prefix_len]
    every_nth = len(words) // 1000
    _words = [w[:prefix_len] for w in words[::every_nth]]
    return _words[:1000]


WORDS100k = words100k()
MIXED_WORDS100k = truncated_words(WORDS100k)
NON_WORDS100k = random_words(100000)
PREFIXES_3_1k = prefixes1k(WORDS100k, 3)
PREFIXES_5_1k = prefixes1k(WORDS100k, 5)
PREFIXES_8_1k = prefixes1k(WORDS100k, 8)
PREFIXES_15_1k = prefixes1k(WORDS100k, 15)


def format_result(key, value, text_width):
    key = key.ljust(text_width)
    print(f"    {key} {value}")


def bench(
    name, timer, descr="M ops/sec", op_count=0.1, repeats=3, runs=5, text_width=33
):
    try:
        times = []
        for x in range(runs):
            times.append(timer.timeit(repeats))

        def op_time(time):
            return op_count * repeats / time

        val = f"{op_time(min(times)):0.3f}{descr}"
        format_result(name, val, text_width)
    except (AttributeError, TypeError):
        format_result(name, "not supported", text_width)


def create_trie():
    words = WORDS100k
    return marisa_trie.Trie(words)


def create_bytes_trie():
    words = WORDS100k
    values = (struct.pack("<H", len(word)) for word in words)
    return marisa_trie.BytesTrie(zip(words, values))


def create_record_trie():
    words = WORDS100k
    values = ([len(word)] for word in words)
    return marisa_trie.RecordTrie("<H", zip(words, values))


def benchmark():
    print("\n====== Benchmarks (100k unique unicode words) =======\n")

    common_setup = """
from __main__ import create_trie, create_bytes_trie, create_record_trie
from __main__ import WORDS100k, NON_WORDS100k, MIXED_WORDS100k
from __main__ import PREFIXES_3_1k, PREFIXES_5_1k, PREFIXES_8_1k, PREFIXES_15_1k
NON_WORDS_10k = NON_WORDS100k[:10000]
NON_WORDS_1k = ['ыва', 'xyz', 'соы', 'Axx', 'avы']*200
"""
    BUILD = {
        "dict": "data = dict((word, len(word)) for word in WORDS100k);",
        "Trie": "data = create_trie();",
        "BytesTrie": "data = create_bytes_trie();",
        "RecordTrie": "data = create_record_trie();",
    }

    dict_setup = common_setup + BUILD["dict"]
    trie_setup = common_setup + BUILD["Trie"]
    bytes_trie_setup = common_setup + BUILD["BytesTrie"]
    record_trie_setup = common_setup + BUILD["RecordTrie"]

    structures = [
        ("dict", dict_setup),
        ("Trie", trie_setup),
        ("BytesTrie", bytes_trie_setup),
        ("RecordTrie", record_trie_setup),
    ]

    # build performance
    for name, setup in structures:
        timer = timeit.Timer(BUILD[name], common_setup)
        bench(f"{name} building", timer, "M words/sec", 0.1, 5)

    # common operations speed
    tests = [
        (
            "__getitem__ (hits)",
            "for word in WORDS100k: data[word]",
            "M ops/sec",
            0.1,
            3,
        ),
        ("get() (hits)", "for word in WORDS100k: data.get(word)", "M ops/sec", 0.1, 3),
        (
            "get() (misses)",
            "for word in NON_WORDS_10k: data.get(word)",
            "M ops/sec",
            0.01,
            5,
        ),
        (
            "__contains__ (hits)",
            "for word in WORDS100k: word in data",
            "M ops/sec",
            0.1,
            3,
        ),
        (
            "__contains__ (misses)",
            "for word in NON_WORDS100k: word in data",
            "M ops/sec",
            0.1,
            3,
        ),
        ("items()", "list(data.items())", " ops/sec", 1, 1),
        ("keys()", "list(data.keys())", " ops/sec", 1, 1),
        #        ('values()', 'list(data.values())', ' ops/sec', 1, 1),
    ]

    for test_name, test, descr, op_count, repeats in tests:
        for name, setup in structures:
            timer = timeit.Timer(test, setup)
            full_test_name = f"{name} {test_name}"
            bench(full_test_name, timer, descr, op_count, repeats, 9)

    # trie-specific benchmarks
    for struct_name, setup in structures[1:]:
        _bench_data = [
            ("hits", "WORDS100k"),
            ("mixed", "MIXED_WORDS100k"),
            ("misses", "NON_WORDS100k"),
        ]

        for meth in ["prefixes", "iter_prefixes"]:
            for name, data in _bench_data:
                bench(
                    f"{struct_name}.{meth} ({name})",
                    timeit.Timer(
                        f"for word in {data}:\n"
                        f"   for it in data.{meth}(word): pass",
                        setup,
                    ),
                    runs=3,
                )

        _bench_data = [
            ("xxx", "avg_len(res)==415", "PREFIXES_3_1k"),
            ("xxxxx", "avg_len(res)==17", "PREFIXES_5_1k"),
            ("xxxxxxxx", "avg_len(res)==3", "PREFIXES_8_1k"),
            ("xxxxx..xx", "avg_len(res)==1.4", "PREFIXES_15_1k"),
            ("xxx", "NON_EXISTING", "NON_WORDS_1k"),
        ]
        for xxx, avg, data in _bench_data:
            for meth in ["keys"]:  # ('items', 'keys', 'values'):
                bench(
                    f'{struct_name}.{meth}(prefix="{xxx}"), {avg}',
                    timeit.Timer(
                        f"for word in {data}: data.{meth}(word)", setup
                    ),
                    "K ops/sec",
                    op_count=1,
                    runs=3,
                    text_width=60,
                )


def check_trie(trie, words):
    assert sum(trie[word] for word in words) == len(words)


def profiling():
    import pstats
    import cProfile

    print("\n====== Profiling =======\n")
    trie = create_trie()
    WORDS = WORDS100k

    #    def check_prefixes(trie, words):
    #        for word in words:
    #            trie.keys(word)
    #    cProfile.runctx("check_prefixes(trie, NON_WORDS_1k)", globals(), locals(), "Profile.prof")
    #
    cProfile.runctx("check_trie(trie, WORDS)", globals(), locals(), "Profile.prof")

    s = pstats.Stats("Profile.prof")
    s.strip_dirs().sort_stats("time").print_stats(20)


if __name__ == "__main__":
    benchmark()
    # profiling()
    print("\n~~~~~~~~~~~~~~\n")



import pickle
from uuid import uuid4

import pytest
import hypothesis.strategies as st
from hypothesis import given, assume

import marisa_trie

from .utils import text, Mapping


@given(st.sets(text), text)
def test_init(keys, missing_key):
    assume(missing_key not in keys)

    trie = marisa_trie.Trie(keys)
    for key in keys:
        assert key in trie

    assert missing_key not in trie


@given(st.sets(text, min_size=1), text)
def test_key_id(keys, missing_key):
    assume(missing_key not in keys)

    trie = marisa_trie.Trie(keys)
    for key in keys:
        key_id = trie.key_id(key)
        assert trie.restore_key(key_id) == key

    key_ids = [trie.key_id(key) for key in keys]
    non_existing_id = max(key_ids) + 1

    with pytest.raises(KeyError):
        trie.restore_key(non_existing_id)

    with pytest.raises(KeyError):
        trie.key_id(missing_key)


@given(st.sets(text, min_size=1), text)
def test_getitem(keys, missing_key):
    assume(missing_key not in keys)

    trie = marisa_trie.Trie(keys)
    for key in keys:
        key_id = trie[key]
        assert trie.restore_key(key_id) == key

    key_ids = [trie[key] for key in keys]
    non_existing_id = max(key_ids) + 1

    with pytest.raises(KeyError):
        trie.restore_key(non_existing_id)

    with pytest.raises(KeyError):
        trie[missing_key]


@given(st.sets(text))
def test_get(keys):
    trie = marisa_trie.Trie(keys)
    for key in keys:
        key_id = trie.get(key)
        assert trie.restore_key(key_id) == key

        key_id = trie.get(key.encode("utf8"))
        assert trie.restore_key(key_id) == key

        key_id = trie.get(key, "default value")
        assert trie.restore_key(key_id) == key

    assert trie.get("non_existing_key") is None
    assert trie.get(b"non_existing_bytes_key") is None
    assert trie.get("non_existing_key", "default value") == "default value"
    assert trie.get(b"non_existing_bytes_key", "default value") == "default value"


@given(st.sets(text))
def test_saveload(tmpdir_factory, keys):
    trie = marisa_trie.Trie(keys)

    dirname = f"{str(uuid4())}_"
    path = str(tmpdir_factory.mktemp(dirname).join("trie.bin"))
    trie.save(path)

    trie2 = marisa_trie.Trie()
    trie2.load(path)

    for key in keys:
        assert key in trie2


@given(st.sets(text))
def test_mmap(tmpdir_factory, keys):
    trie = marisa_trie.Trie(keys)

    dirname = f"{str(uuid4())}_"
    path = str(tmpdir_factory.mktemp(dirname).join("trie.bin"))
    trie.save(path)

    trie2 = marisa_trie.Trie()
    trie2.mmap(path)

    for key in keys:
        assert key in trie2

@given(st.sets(text))
def test_map(tmpdir_factory, keys):
    trie = marisa_trie.Trie(keys)

    dirname = f"{str(uuid4())}_"
    path = str(tmpdir_factory.mktemp(dirname).join("trie.bin"))
    trie.save(path)

    data = open(path, "rb").read()
    trie2 = marisa_trie.Trie()
    trie2.map(data)

    for key in keys:
        assert key in trie2

@given(st.sets(text))
def test_map_with_pad(tmpdir_factory, keys):
    trie = marisa_trie.Trie(keys)

    dirname = f"{str(uuid4())}_"
    path = str(tmpdir_factory.mktemp(dirname).join("trie.bin"))
    trie.save(path)

    data = b"pad" + open(path, "rb").read() + b"pad"
    trie2 = marisa_trie.Trie()
    trie2.map(memoryview(data)[3:-3])

    for key in keys:
        assert key in trie2


@given(st.sets(text))
def test_tobytes_frombytes(keys):
    trie = marisa_trie.Trie(keys)
    data = trie.tobytes()

    trie2 = marisa_trie.Trie().frombytes(data)

    for key in keys:
        assert key in trie2
        assert trie2.key_id(key) == trie.key_id(key)


@given(st.sets(text))
def test_dumps_loads(keys):
    trie = marisa_trie.Trie(keys)
    data = pickle.dumps(trie)

    trie2 = pickle.loads(data)

    for key in keys:
        assert key in trie2
        assert trie2.key_id(key) == trie.key_id(key)


def test_contains_empty():
    assert "foo" not in marisa_trie.Trie()


def test_contains_singleton():
    trie = marisa_trie.Trie(["foo"])
    assert "foo" in trie
    assert "f" not in trie


def test_eq_self():
    trie = marisa_trie.Trie()
    assert trie == trie
    assert trie == marisa_trie.Trie()


def test_eq_neq():
    trie = marisa_trie.Trie(["foo", "bar"])
    assert trie == marisa_trie.Trie(["foo", "bar"])
    assert trie != marisa_trie.Trie(["foo", "boo"])


def test_neq_different_type():
    assert marisa_trie.Trie(["foo", "bar"]) != {}


def test_eq_neq_different_order():
    lo_trie = marisa_trie.Trie(order=marisa_trie.LABEL_ORDER)
    wo_trie = marisa_trie.Trie(order=marisa_trie.WEIGHT_ORDER)
    assert lo_trie == lo_trie and wo_trie == wo_trie
    assert lo_trie != wo_trie


def test_gt_lt_exceptions():
    with pytest.raises(TypeError):
        marisa_trie.Trie() < marisa_trie.Trie()

    with pytest.raises(TypeError):
        marisa_trie.Trie() > marisa_trie.Trie()


def test_iter():
    trie = marisa_trie.Trie(["foo", "bar"])
    assert list(trie) == list(trie.iterkeys())


def test_len():
    trie = marisa_trie.Trie()
    assert len(trie) == 0

    trie = marisa_trie.Trie(["foo", "f", "bar"])
    assert len(trie) == 3


def test_prefixes():
    trie = marisa_trie.Trie(["foo", "f", "foobar", "bar"])
    assert trie.prefixes("foobar") == ["f", "foo", "foobar"]
    assert trie.prefixes("foo") == ["f", "foo"]
    assert trie.prefixes("bar") == ["bar"]
    assert trie.prefixes("b") == []

    assert list(trie.iter_prefixes("foobar")) == ["f", "foo", "foobar"]

def test_iter_prefixes_with_keys():
    trie = marisa_trie.Trie(["foo", "f", "foobar", "bar"])

    assert set(trie.iter_prefixes_with_ids("foobar")) == {
        ("f", trie["f"]),
        ("foo", trie["foo"]),
        ("foobar", trie["foobar"]),
    }
    assert set(trie.iter_prefixes_with_ids("foo")) == {
        ("f", trie["f"]),
        ("foo", trie["foo"]),
    }
    assert set(trie.iter_prefixes_with_ids("bar")) == {("bar", trie["bar"])}
    assert not set(trie.iter_prefixes_with_ids("b"))

    for test_key in ["foobar", "foo", "bar", "b"]:
        assert list(trie.iter_prefixes_with_ids(test_key)) == [
            (prefix, trie[prefix]) for prefix in trie.prefixes(test_key)
        ]

def test_keys():
    keys = ["foo", "f", "foobar", "bar"]
    trie = marisa_trie.Trie(keys)
    assert set(trie.keys()) == set(keys)


def test_keys_prefix():
    keys = ["foo", "f", "foobar", "bar"]
    trie = marisa_trie.Trie(keys)
    assert set(trie.keys("fo")) == {"foo", "foobar"}
    assert trie.keys("foobarz") == []


@given(st.sets(text))
def test_iterkeys(keys):
    trie = marisa_trie.Trie(keys)
    assert trie.keys() == list(trie.iterkeys())

    for key in keys:
        prefix = key[:5]
        assert trie.keys(prefix) == list(trie.iterkeys(prefix))


def test_items():
    keys = ["foo", "f", "foobar", "bar"]
    trie = marisa_trie.Trie(keys)
    items = trie.items()
    assert set(items) == set(zip(keys, (trie[k] for k in keys)))


def test_items_prefix():
    keys = ["foo", "f", "foobar", "bar"]
    trie = marisa_trie.Trie(keys)
    assert set(trie.items("fo")) == {
        ("foo", trie["foo"]),
        ("foobar", trie["foobar"]),
    }


@given(st.sets(text))
def test_iteritems(keys):
    trie = marisa_trie.Trie(keys)
    assert trie.items() == list(trie.iteritems())

    for key in keys:
        prefix = key[:5]
        assert trie.items(prefix) == list(trie.iteritems(prefix))


@pytest.mark.filterwarnings("ignore:Trie.has_keys_with_prefix is deprecated")
def test_has_keys_with_prefix_empty():
    empty_trie = marisa_trie.Trie()
    assert not empty_trie.has_keys_with_prefix("")
    assert not empty_trie.has_keys_with_prefix("ab")


@pytest.mark.filterwarnings("ignore:Trie.has_keys_with_prefix is deprecated")
def test_has_keys_with_prefix():
    fruit_trie = marisa_trie.BytesTrie(
        [
            ("apple", b"foo"),
            ("pear", b"bar"),
            ("peach", b"baz"),
        ]
    )
    assert fruit_trie.has_keys_with_prefix("")
    assert fruit_trie.has_keys_with_prefix("a")
    assert fruit_trie.has_keys_with_prefix("pe")
    assert fruit_trie.has_keys_with_prefix("pear")
    assert not fruit_trie.has_keys_with_prefix("x")


def test_invalid_file():
    try:
        marisa_trie.Trie().load(__file__)
    except RuntimeError as e:
        assert "MARISA_FORMAT_ERROR" in e.args[0]
    else:
        pytest.fail("Exception is not raised")


def test_mutable_mapping():
    for method in Mapping.__abstractmethods__:
        assert hasattr(marisa_trie.Trie, method)



import io
import pickle

import hypothesis.strategies as st
from hypothesis import given

import marisa_trie

from .utils import text

records = st.tuples(st.integers(min_value=0, max_value=2 ** 16 - 1), st.booleans())


@given(st.sets(st.tuples(text, records)))
def test_dumps_loads(data):
    trie = marisa_trie.RecordTrie("<H?", data)

    buf = io.BytesIO()
    pickle.dump(trie, buf)
    buf.seek(0)

    assert trie == pickle.load(buf)



import itertools
from pathlib import Path

from Cython.Build import cythonize
from setuptools import setup, Extension


MARISA_ROOT_DIR = Path("marisa-trie")
MARISA_SOURCE_DIR = MARISA_ROOT_DIR / "lib"
MARISA_INCLUDE_DIR = MARISA_ROOT_DIR / "include"
MARISA_FILES = [
    "marisa/*.cc",
    "marisa/grimoire.cc",
    "marisa/grimoire/io/*.cc",
    "marisa/grimoire/trie/*.cc",
    "marisa/grimoire/vector/*.cc",
]

MARISA_FILES[:] = map(str, itertools.chain(
        *(MARISA_SOURCE_DIR.glob(path) for path in MARISA_FILES)
    )
)

extensions = [
    Extension(
        "marisa_trie",
        sources=["src/*.pyx"],
        language="c++",
        include_dirs=[str(MARISA_INCLUDE_DIR)],
    ),
]

setup(
    libraries=[
        (
            "libmarisa-trie",
            {
                "sources": MARISA_FILES,
                "include_dirs": [str(MARISA_SOURCE_DIR), str(MARISA_INCLUDE_DIR)],
            },
        )
    ],
    ext_modules=cythonize(extensions, language_level="3"),
)
